﻿/**
 * ...
 * @author Georg Graf
 */

package com.georg
{
	import com.georg.tools.ComplexNumber;
	import com.georg.tools.RGB;
	
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.TimerEvent;
	import flash.geom.ColorTransform;
	import flash.utils.Timer;
	
	public class Mandelbrot extends Sprite
	{
		protected var bm:Bitmap;
		
		// draw preferences
		public var colorTable:Array;
		public var colorAngle:int;
		public var iterations:int;
		public var scale:Number;
		public var offsetX:Number;
		public var offsetY:Number;
		
		// drawIncremental variables
		private var timer:Timer;
		private var tileSteps:int;
		private var tileWidth:int;
		private var tileHeight:int;
		private var tileStartX:int;
		private var tileStartY:int;
		private var onCompleteFunction:Function;
		
		// drawPixelLike variables
		private var pixSteps:int;
		private var pixTimer:Timer;
		private var pixWidth:int;
		private var pixHeight:int;
		private var pixStartX:int;
		private var pixStartY:int;
		
		// draw variables
		private var cnZ:ComplexNumber;
		private var count:int;
		
		public function Mandelbrot(width:int, height:int) 
		{
			offsetX = 0;
			offsetY = 0;
			
			bm = new Bitmap(new BitmapData(width, height, false, 0));
			addChild(bm);
		}
		
		public function drawIncremental(steps:int = 4, delay:Number = 10, iterations:int = 20, scale:Number = .01, colorAngle:int = 10, onComplete:Function = null):void
		{
			this.iterations = iterations;
			this.scale = scale;
			this.colorAngle = colorAngle;
			this.onCompleteFunction = onComplete;
			
			if (timer != null)
			{
				timer.stop();
				timer.removeEventListener(TimerEvent.TIMER, drawPiece);
				timer = null;
			}
			
			createColorTable();
			
			// set calculation variables
			tileSteps = steps;
			tileWidth = width / tileSteps;
			tileHeight = height / tileSteps;
			
			// start timing
			timer = new Timer(delay, steps * steps);
			timer.addEventListener(TimerEvent.TIMER, drawPiece);
			timer.start();
		}
		
		public function drawPixelIncremental(steps:int = 10, delay:Number = 10, iterations:int = 20, scale:Number = .01, colorAngle:int = 10, onComplete:Function = null):void
		{
			this.iterations = iterations;
			this.scale = scale;
			this.colorAngle = colorAngle;
			this.onCompleteFunction = onComplete;
			
			if (pixTimer != null)
			{
				pixTimer.stop();
				pixTimer.removeEventListener(TimerEvent.TIMER, drawPiece);
				pixTimer = null;
			}
			
			createColorTable();
			
			// set calculation variables
			pixSteps = steps;
			pixWidth = width / pixSteps;
			pixHeight = height / pixSteps;
			
			// start timing
			pixTimer = new Timer(delay, pixWidth * pixHeight);
			pixTimer.addEventListener(TimerEvent.TIMER, drawPixelLike);
			pixTimer.start();
		}
		
		public function draw(iterations:int = 20, scale:Number = .01, colorAngle:int = 10):void
		{
			this.iterations = iterations;
			this.scale = scale;
			this.colorAngle = colorAngle;
			
			createColorTable();
			drawSegment(-(width / 2), 	// -half width
						(width / 2), 	// +half width
						-(height / 2), 	// -half height
						(height / 2));	// +half height
		}
		
		
		public function drag(offX:Number, offY:Number):void
		{
			if (timer != null)		timer.stop();
			if (pixTimer != null)	pixTimer.stop();
			
			offsetX -= offX;
			offsetY -= offY;
			
			bm.bitmapData.scroll(offX, offY);
		}
		
		public function getIterationsForPoint(pX:int, pY:int):Array
		{
			var arr:Array = new Array();
			cnZ = new ComplexNumber(0, 0);
			var cnt:int = 0;
			
			arr.push(cnZ);
			
			while(cnt < iterations)
			{
				// z(n+1) = z(n)^2 + c
				cnZ = (cnZ.power()).add(new ComplexNumber((pX + offsetX) * scale, (pY + offsetY) * scale));
				cnt++;

				arr.push(cnZ.clone());
				
				if (cnZ.abs() > 2)
					break;
			}
			
			return arr;
		}
	
		private function drawPiece(e:TimerEvent):void
		{
			tileStartX = (((timer.currentCount - 1) % tileSteps) * tileWidth) - (width / 2);
			tileStartY = (Math.floor((timer.currentCount - 1) / tileSteps) * tileHeight) - (height / 2);
			
			drawSegment(tileStartX, 
						tileStartX + tileWidth,
						tileStartY,
						tileStartY + tileHeight);
						
			if (timer.currentCount == timer.repeatCount)
				onCompleteFunction();
		}
		
		private function drawPixelLike(e:TimerEvent):void
		{
			pixStartX = int( (((pixTimer.currentCount - 1) % pixWidth)) - (width / 2));
			pixStartY = int((Math.floor((pixTimer.currentCount - 1) / pixWidth)) - (height / 2));
			
			for (var iX:int = 0; iX < pixSteps; iX++)
				for (var iY:int = 0; iY < pixSteps; iY++)
					drawPixel(pixStartX + (iX * pixWidth), pixStartY + (iY * pixHeight));
		}
		
		private function createColorTable():void
		{
			colorTable = new Array();
					
			for (var i:int = 0; i < iterations; i++)
			//trace(RGB.hsv(10, 1, (i / iterations * .5)));
				colorTable.push(RGB.hsv(colorAngle, 1, (i / iterations)));
		}
		
		/** the core-function to draw a segment of the Mandelbrot **/
		private function drawSegment(startX:int, endX:int, startY:int, endY:int):void
		{
			// create image 
			for (var iX:int = startX; iX < endX; iX++)
				for (var iY:int = startY; iY < endY; iY++)
					drawPixel(iX, iY);
		}
		
		/** the core-function to draw one pixel of the Mandelbrot **/
		private function drawPixel(posX:int, posY:int):void
		{
			cnZ = new ComplexNumber(0, 0);
			count = 0;
			
			while(count < iterations)
			{
				// z(n+1) = z(n)^2 + c
				cnZ = (cnZ.power()).add(new ComplexNumber((posX + offsetX) * scale, (posY + offsetY) * scale));
				count++;
				
				if (cnZ.abs() > 2)
					break;
			}
			bm.bitmapData.setPixel(posX + (width / 2), posY + (height / 2), colorTable[count]);
		}
	}
}